#include <stdlib.h>

#include "lua.h"
#include "lauxlib.h"

#include "xmlparse.h"

static int parser_tag;

struct lxp_userdata {
    lua_State *L;
    XML_Parser *parser;
    int tableref;
};

typedef struct lxp_userdata lxp_userdata;

        
static void startElement(void *userData, const char *name, const char **attrs)
{
    lxp_userdata *l_userdata = userData;
    lua_State *L = l_userdata->L;
    const char **p;

    lua_getref(L, l_userdata->tableref);
    lua_pushstring(L, "starttag");
    lua_gettable(L, -2);
    if (lua_isnil(L, -1)) {
        return;
    }
    if (!lua_isfunction(L, -1)) {
        lua_error(L, "xml starttag callback is not a function");
    }

    lua_pushusertag(L, l_userdata->parser, parser_tag);
    lua_pushstring(L, name);
    // lua_pushnil(L);

    lua_newtable(L);
    p = attrs;
    while (*p) {
        lua_pushstring(L, *p++);
        lua_pushstring(L, *p++);
        lua_settable(L, -3);
    }

    lua_call(L, 3, 0);

    lua_pop(L, 1);
}


static void endElement(void *userData, const char *name)
{
    lxp_userdata *l_userdata = userData;
    lua_State *L = l_userdata->L;

    lua_getref(L, l_userdata->tableref);
    lua_pushstring(L, "endtag");
    lua_gettable(L, -2);
    if (lua_isnil(L, -1)) {
        return;
    }
    if (!lua_isfunction(L, -1)) {
        lua_error(L, "xml endtag callback is not a function");
    }

    lua_pushusertag(L, l_userdata->parser, parser_tag);
    lua_pushstring(L, name);
    lua_call(L, 2, 0);

    lua_pop(L, 1);
}

static void charData(void *userData, const char *s, int len)
{
    lxp_userdata *l_userdata = userData;
    lua_State *L = l_userdata->L;

    lua_getref(L, l_userdata->tableref);
    lua_pushstring(L, "chardata");
    lua_gettable(L, -2);
    if (lua_isnil(L, -1)) {
        return;
    }
    if (!lua_isfunction(L, -1)) {
        lua_error(L, "xml chardata callback is not a function");
    }

    lua_pushusertag(L, l_userdata->parser, parser_tag);
    lua_pushlstring(L, s, len);
    lua_call(L, 2, 0);

    lua_pop(L, 1);
}    


static int lxp_make_parser (lua_State *L) {
    XML_Parser parser = XML_ParserCreate(NULL);
    lxp_userdata *l_userdata = malloc(sizeof(lxp_userdata));

    if (!l_userdata) {
        lua_error(L, "lxp_userdata malloc failed");
    }

    if (!parser) {
        lua_error(L, "XML_ParserCreate failed");
    }

    l_userdata->L = L;
    l_userdata->parser = parser;
    lua_newtable(L);
    l_userdata->tableref = lua_ref(L, 1);

    XML_SetUserData(parser, l_userdata);

    XML_SetElementHandler(parser, startElement, endElement);
    XML_SetCharacterDataHandler(parser, charData);
    
    lua_pushusertag(L, parser, parser_tag);
    return 1;
}


static int parser_gc (lua_State *L) {
  XML_Parser parser = lua_touserdata(L, 1);
  lxp_userdata *l_userdata = XML_GetUserData(parser);

  lua_unref(L, l_userdata->tableref);
  free(l_userdata);

  XML_ParserFree(parser);
  return 0;
}


/* pass through table references to the XML_UserData table inside */

static int parser_gettable (lua_State *L) {
  XML_Parser parser = lua_touserdata(L, 1);
  lxp_userdata *l_userdata = XML_GetUserData(parser);
  lua_getref(L, l_userdata->tableref);

  /* shuffle the ref to the bottom of the stack so we can just call the normal gettable */
  lua_remove(L, 1);
  lua_insert(L, 1);
  lua_gettable(L, 1);
  return 1;
}

static int parser_settable (lua_State *L) {
  XML_Parser parser = lua_touserdata(L, 1);

  lxp_userdata *l_userdata = XML_GetUserData(parser);
  lua_getref(L, l_userdata->tableref);

  /* shuffle the ref to the bottom of the stack so we can just call the normal settable */
  lua_remove(L, 1);
  lua_insert(L, 1);
  lua_settable(L, 1);
  return 0;
}


static int lxp_parse(lua_State *L) {
    XML_Parser parser = lua_touserdata(L, 1);
    const char *s;
    int len;

    if (!parser) {
        lua_error(L, "parse arg must be Parser userdata");
    }
    
    s = luaL_check_lstr(L, 2, &len);
    
    if (XML_Parse(parser, s, len, 0)) {
      lua_pushnumber(L, 1);
    } else {
      /* lua_error(L, "parse error"); */
      lua_pushnil(L);
    }
    return 1;
}

static int lxp_parse_end(lua_State *L) {
    XML_Parser parser = lua_touserdata(L, 1);

    if (!parser) {
        lua_error(L, "parse arg must be Parser userdata");
    }
    
    if (XML_Parse(parser, "", 0, 1)) {
      lua_pushnumber(L, 1);
    } else {
      /* lua_error(L, "parse error"); */
      lua_pushnil(L);
    }
    return 1;
}

static int lxp_error_code(lua_State *L) {
    XML_Parser parser = lua_touserdata(L, 1);

    if (!parser) {
        lua_error(L, "parse arg must be Parser userdata");
    }

    lua_pushnumber(L, XML_GetErrorCode(parser));
    return 1;
}

static int lxp_error_string(lua_State *L) {
    XML_Parser parser = lua_touserdata(L, 1);

    if (!parser) {
        lua_error(L, "parse arg must be Parser userdata");
    }

    lua_pushstring(L, XML_ErrorString(XML_GetErrorCode(parser)));
    return 1;
}

static int lxp_get_line_number(lua_State *L) {
    XML_Parser parser = lua_touserdata(L, 1);

    if (!parser) {
        lua_error(L, "parse arg must be Parser userdata");
    }

    lua_pushnumber(L, XML_GetCurrentLineNumber(parser));
    return 1;
}

static int lxp_get_column_number(lua_State *L) {
    XML_Parser parser = lua_touserdata(L, 1);

    if (!parser) {
        lua_error(L, "parse arg must be Parser userdata");
    }

    lua_pushnumber(L, XML_GetCurrentColumnNumber(parser));
    return 1;
}


static const struct luaL_reg lxp_funcs[] = {
    {"lxp_make_parser", lxp_make_parser},
    {"lxp_parse", lxp_parse},
    {"lxp_parse_end", lxp_parse_end},
    {"lxp_error_code", lxp_error_code},
    {"lxp_error_string", lxp_error_string},
    {"lxp_get_line_number", lxp_get_line_number},
    {"lxp_get_column_number", lxp_get_column_number},
};

void lua_lxplibopen(lua_State *L) {
    luaL_openl(L, lxp_funcs);

    parser_tag = lua_newtag(L);
    lua_pushcfunction(L, parser_gc);
    lua_settagmethod(L, parser_tag, "gc");

    lua_pushcfunction(L, parser_gettable);
    lua_settagmethod(L, parser_tag, "gettable");

    lua_pushcfunction(L, parser_settable);
    lua_settagmethod(L, parser_tag, "settable");
}
